{{ DCC Controller }}
CON

   _clkmode = xtal1 + pll16x                             ' use crystal x 16
   _xinfreq = 5_000_000                                  ' 5 MHz cyrstal (sys clock = 80 MHz)

   clockfreq = ((_CLKMODE - XTAL1) >> 6) * _XINFREQ

   _1uS  = clockfreq / 1_000_000 'Divisor for  1 uS
   _10uS = clockfreq /   100_000 'Divisor for 10 uS
   _1mS  = clockfreq /     1_000 'Divisor for  1 mS
   _1S   = clockfreq /         1 'Divisor for  1 S

   FAST_CLOCK = (clockfreq / 60) * 60

   LCD_PIN   = 0                                         ' for ModernDevice 4x20 serial LCD on P0
   LCD_BAUD  = 9600
   LCD_LINES = 4

   DCC_PIN = 1
   BRAKE_PIN = 2
   STROBE_PIN = 3
   SENSE_PIN  = 4
   FEEDBACK_PIN = 5

   {{
     Booster pinout:
          curr_sense = blue
          +5         = red
          dcc        = yellow
          brake      = green
          GND        = black
   }}

   NTRAINS = 4
   TRAIN_ADDRESS   = 0
   TRAIN_SPEED     = 1
   TRAIN_FUNCTIONS = 2
   TRAIN_CONFIG    = 3
   TRAIN_NENTRIES  = 4

    CONFIG_ADDRESS = 0                                      ' if this bit is set, use 4 digit addresses, else 2 digits

    PREAMBLE       = 16

    BROADCAST_ADDR = 0

    ENTER_KEY      = $0D
    ESC_KEY        = $CB
    BS_KEY         = $C8
    DEL_KEY        = $C9

OBJ
  lcd : "LCD_driver"   {{ driver for ModernDevices serial 4x20 LCD }}
  dcc : "DCC_Driver"
  kbd : "keyboard"


VAR
  'LONG frames[406] '4 cogs measure 1600 samples up to 80mhz
  LONG stack2[32]
  LONG stack4[32]
  LONG stack5[32]
  LONG stack3[64]

  LONG trains[TRAIN_NENTRIES*NTRAINS]

  LONG sel_train ' currently selected train. Commands will act upon this train
  LONG sense_state

  BYTE hour, ampm, minute, tick
  LONG dcc_cog

PUB main | ok, key, tmp, train         {{ COG: Main entry point }}
   ' start the LCD driver
   ok := lcd.init(LCD_PIN, LCD_BAUD)

   ' start the DCC packet driver
   ok := dcc.start(@DCCBuffer, DCC_PIN, STROBE_PIN)

   'start the keyboard
   kbd.startx(26, 27, %0_000_100, %01_11111)

   DIRA[23..16]~~

   ' initialize the trains arrays
   repeat train from 0 to NTRAINS-1
      trains[train*TRAIN_NENTRIES + TRAIN_ADDRESS] := LOOKUPZ(train : 1146, 8952, 7120, 8956)
      trains[train*TRAIN_NENTRIES + TRAIN_SPEED] := 0
      trains[train*TRAIN_NENTRIES + TRAIN_FUNCTIONS] := 0
      trains[train*TRAIN_NENTRIES + TRAIN_CONFIG] := LOOKUPZ(train : 1, 1, 1, 1)
   sel_train := 0

   repeat 4
      waitcnt(clkfreq+cnt) ' wait couple of secs for the LCD to fully power up
   lcd.cursor(0)
   Welcome

   cognew(FastClock,@stack2)
   cognew(FeedbackChecker, @stack4)
   cognew(FeedbackLED, @stack5)
   OperationMode(1)

   ' initialize the fast clock
   hour   := 6
   minute := 0
   ampm   := 0

   waitcnt(_1mS+cnt)    ' wait for system to finish initializing
   DIRA[BRAKE_PIN]:=1
   OUTA[BRAKE_PIN]:=0   ' turn off brake, enabling DCC output

   ' initialize the display for all trains
   repeat
      RefreshTrainStatus
      sense_state := 0
      key := kbd.key
      case key
         "1".."4":
            sel_train := key - "1"
         194, 450:  ' up arrow, shift up arrow
            tmp := trains[sel_train*TRAIN_NENTRIES+TRAIN_SPEED]
            if key == 194
               tmp := tmp + 1 <# 127
            else
               tmp := tmp + 10 <# 127
            trains[sel_train*TRAIN_NENTRIES+TRAIN_SPEED] := tmp
         195, 451:  ' down arrow, shift down arrow
            tmp := trains[sel_train*TRAIN_NENTRIES+TRAIN_SPEED]
            if key == 195
               tmp := tmp - 1 #> -127
            else
               tmp := tmp - 10 #> -127
            trains[sel_train*TRAIN_NENTRIES+TRAIN_SPEED] := tmp
         "0": ' F0
            trains[sel_train*TRAIN_NENTRIES+TRAIN_FUNCTIONS] ^= 1
         208..217: ' F1 .. F10
            tmp := key - 208 + 1 ' bit for the function in question
            tmp := 1 << tmp
            trains[sel_train*TRAIN_NENTRIES+TRAIN_FUNCTIONS] ^= tmp
         464..472: ' shift F1 .. F9 (ie F11-F19)
            tmp := key - 464 + 11 ' bit for the function in question
            tmp := 1 << tmp
            trains[sel_train*TRAIN_NENTRIES+TRAIN_FUNCTIONS] ^= tmp
         32: ' space: emergency stop for all trains
            repeat tmp from 0 to NTRAINS-1
               trains[tmp*TRAIN_NENTRIES+TRAIN_SPEED] := 0
         ESC_KEY: ' escape: emergency stop for sel train
            trains[sel_train*TRAIN_NENTRIES+TRAIN_SPEED] := 0
         "C": ' enter configuration mode
            ConfigurationMode
            OperationMode(0)
         "P": ' enter programming mode
            ExitOpMode
            ProgrammingMode
            OperationMode(1)
         0:  ' no key: do nothing
         other:
            lcd.to_xy(15,3)
            lcd.dec(key)
            lcd.str(@spaces_str)
            lcd.wait100ms


PUB ReadInt(limit) | key, val, newval  {{ read a positive integer, but no larger than limit }}
    lcd.cursor(3)
    val := 0
    repeat
          key := kbd.getkey
          case key
               "0".."9":
                   newval  := val * 10 + key - "0"
                   if newval =< limit
                      val := newval
                      lcd.out(key)
               13:
                    lcd.cursor(0)
                    return val
               BS_KEY, DEL_KEY:
                   if val > 0
                      lcd.backspace
                      val := val / 10
               ESC_KEY:
                   repeat while val > 0
                      lcd.backspace
                      val := val / 10

PUB ConfigurationMode | key            {{ Configures the controller: address of each train etc }}
   ' for each train, configure: address, address type (7,14), speed type (28, 128)
   ' TO BE DONE
   lcd.clear
   lcd.to_xy(0,0)
   lcd.str(@config_str)
   lcd.to_xy(0,3)
   lcd.str(@any_key_str)
   key := kbd.getkey

PUB ProgrammingMode | key,cv, cvs, val {{ Programming mode to configure decoders }}
   ''' enter service mode
   repeat 3
      SendDCCReset
   repeat
      'OUTA[BRAKE_PIN]:=1   ' turn off power
      sense_state := 0
      lcd.clear
      lcd.to_xy(0,0)
      lcd.str(@progrm_str)
      lcd.to_xy(0,1)
      lcd.str(@variable_str)
      repeat
            sense_state := 0
            cv := ReadInt(1023)
      while cv == 0
      lcd.to_xy(0,2)
      lcd.str(@prg_act_str)
      repeat
         sense_state := 0
         key := kbd.getkey
      until ((key == "X") OR (key == "S") OR (key == "Q"))
      lcd.out(key)
      case key
         "X":
             ' exit service mode
             return
         "S":
             'set a config variable
             lcd.to_xy(0,1)
             lcd.clearline
             lcd.str(@value_str)
             lcd.to_xy(4,1)
             lcd.dec(cv)
             lcd.to_xy(16,1)
             val := ReadInt(255)
             lcd.to_xy(0,2)
             lcd.clearline
             lcd.str(@prgw_act_str)
             repeat
                   key := kbd.getkey
             until ((key == "X") OR (key == "C") OR (key == "W"))
             lcd.out(key)
             if key == "X"
                 return
             if key == "W"
                 cvs := cv - 1
                 '' send the service mode write packet series
                 'OUTA[BRAKE_PIN]:=0   ' turn off brake, effectively turns on power
                 repeat 5
                    SendDCCReset
                 sense_state := 0
                 repeat 6
                    ' directCV addressing packet
                    SendDCC7PacketD(%0111_00_00 | %0000_11_00 | ((cvs >> 8) & %11), cvs & $FF, val)
                    if sense_state == 1
                       QUIT
                 if sense_state == 1
                     repeat 6
                         SendDCCReset
                 'OUTA[BRAKE_PIN]:=1   ' turn on brake, so turning off power to rails
         "Q":
             'query value of config variable
             lcd.to_xy(0,1)
             lcd.clearline
             lcd.str(@value_str)
             lcd.to_xy(4,1)
             lcd.dec(cv)
             'OUTA[BRAKE_PIN] := 0
             cvs := cv - 1
             repeat val from 0 to 255
                 lcd.to_xy(16,1)
                 lcd.dec(val)
                 repeat 5
                    SendDCCReset
                 sense_state:=0
                 repeat 6
                    ' directCV addressing packet
                    SendDCC7PacketD(%0111_00_00 | %0000_01_00 | ((cvs >> 8) & %11), cvs & $FF, val)
                    if sense_state == 1
                       QUIT
                 if sense_state == 1
                     SendDCCReset
                     QUIT
             'OUTA[BRAKE_PIN]:=1   ' turn on brake, turning off power
      lcd.to_xy(0,3)
      lcd.str(@any_key_str)
      key := kbd.getkey

PUB FeedbackChecker | when             {{ monitors the sense pin and whenever it drops makes the sense_state var 1 }}
   sense_state := 0
   repeat
       waitpne(|<SENSE_PIN,|<SENSE_PIN,0)
       sense_state := 1

PUB FeedbackLED
   DIRA[FEEDBACK_PIN]:=1
   repeat
      OUTA[FEEDBACK_PIN]:=sense_state

PUB ExitOpMode | train                 {{ exit operation mode, stop all trains, stop all functions }}
   repeat train from 0 to NTRAINS-1
      trains[train*TRAIN_NENTRIES + TRAIN_SPEED] := 0
      trains[train*TRAIN_NENTRIES + TRAIN_FUNCTIONS] := 0
   waitcnt(clockfreq+cnt) ' wait 1 sec to ensure all settings "took"
   cogstop(dcc_cog)       ' stop sending operation dcc packets
   'OUTA[BRAKE_PIN]:=1   ' turn on brake, effectively turns off power

PUB OperationMode(start)               {{ enter operation mode }}
   lcd.clear
   lcd.to_xy(11,2)
   lcd.str(@op1_str)
   lcd.to_xy(16,3)
   lcd.str(@op2_str)
   ' restart the op mode dcc cog
   if start
      dcc_cog := cognew(DccLoop, @stack3)
   OUTA[BRAKE_PIN]:=0   ' turn off brake, effectively turns on power

PUB FastClock | tmp                    {{ COG: Maintain an accelerated clock for scale time }}
   repeat
      waitcnt(FAST_CLOCK+cnt)
      tmp := minute
      minute := (minute + 1) // 60
      if minute < tmp
         hour += 1
         if hour > 12
            hour -= 12
         if hour == 12
            not ampm

PUB RefreshTrainStatus | train, base, tmp   {{ Refresh LCD Display }}
   repeat train from 0 to NTRAINS-1
      lcd.to_xy(0,train)
      base := train * TRAIN_NENTRIES
      if train == sel_train
         lcd.out(126)
         OUTA[23..16] := trains[base+TRAIN_FUNCTIONS] & $FF
      else
         lcd.out(" ")
      lcd.dec_rj(trains[base+TRAIN_ADDRESS], 4)
      if trains[base+TRAIN_CONFIG]
         lcd.out(":") ' 4 digit address mode
      else
         lcd.out(".") ' 2 digit address mode
      tmp := trains[base+TRAIN_SPEED]
      if tmp => 0
         lcd.out("+")
      else
         lcd.out("-")
         tmp := -tmp
      lcd.dec_rj(tmp, 3)
      lcd.wait100ms
   lcd.to_xy(13,0)
   if hour < 10
      lcd.out("0")
   lcd.dec(hour)
   if tick
      lcd.out(":")
   else
      lcd.out(" ")
   not tick
   if minute < 10
      lcd.out("0")
   lcd.dec(minute)
   if ampm
      lcd.str(@pm_str)
   else
      lcd.str(@am_str)
   'lcd.wait100ms

PUB Welcome                            {{ Display welcome screen on power up }}
   lcd.to_xy(0,0)
   lcd.str(@welcome1_str)
   lcd.nl
   lcd.str(@welcome2_str)
   lcd.nl
   lcd.str(@welcome3_str)
   lcd.nl
   lcd.str(@welcome4_str)
   kbd.newkey

PUB SendDCCReset
   DCCBuffer[1] := PREAMBLE
   DCCBuffer[2] := 0
   DCCBuffer[3] := 0
   DCCBuffer[4] := 0
   DCCBuffer[0] := 3 ' reset packet is 3 bytes
   repeat until DCCBuffer[0] == 0

PUB SendDCC7Packet(address7, cmd)                   {{Command with 7 bit address}}
   DCCBuffer[1]:=PREAMBLE  'set preamble length
   DCCBuffer[2]:=address7 & %0111_1111            ' address 7 bits
   DCCBuffer[3]:=cmd                              ' command
   DCCBuffer[4]:=DCCBUffer[2]^DCCBuffer[3]        ' error check
   DCCBuffer[0]:=3    'finally, set the length of the packet
   repeat until DCCBuffer[0]==0  'loop until packet is grabbed by driver

PUB SendDCC7PacketD(address7, cmd, data)            {{Command with 7 bit address + data}}
   DCCBuffer[1]:=PREAMBLE  'set preamble length
   DCCBuffer[2]:=address7 & %0111_1111            ' address 7 bits
   DCCBuffer[3]:=cmd                              ' command
   DCCBuffer[4]:=data
   DCCBuffer[5]:=DCCBUffer[2]^DCCBuffer[3]^DCCBuffer[4] ' error check
   DCCBuffer[0]:=4    'finally, set the length of the packet
   repeat until DCCBuffer[0]==0  'loop until packet is grabbed by driver

PUB SendDCC7PacketDD(address7, cmd, data1, data2)   {{Command with 7 bit address + 2 data bytes}}
   DCCBuffer[1]:=PREAMBLE  'set preamble length
   DCCBuffer[2]:=address7 & %0111_1111            ' address 7 bits
   DCCBuffer[3]:=cmd                              ' command
   DCCBuffer[4]:=data1
   DCCBuffer[5]:=data2
   DCCBuffer[6]:=DCCBuffer[2]^DCCBuffer[3]^DCCBuffer[4]^DCCBuffer[5] ' error check
   DCCBuffer[0]:=5    'finally, set the length of the packet
   repeat until DCCBuffer[0]==0  'loop until packet is grabbed by driver

PUB SendDCC14Packet(address14, cmd)                 {{Command with 14 bit address}}
   DCCBuffer[1]:=PREAMBLE  'set preamble length
   DCCBuffer[2]:=%11_000000 | ((address14 >> 8) & %0011_1111)         ' address msb 6 bits
   DCCBuffer[3]:=(address14 & $FF)                            ' address lsb 8 bits
   DCCBuffer[4]:=cmd                                        ' command
   DCCBuffer[5]:=DCCBUffer[2]^DCCBuffer[3]^DCCBuffer[4]      ' error check
   DCCBuffer[0]:=4    'finally, set the length of the packet
   repeat until DCCBuffer[0]==0  'loop until packet is grabbed by driver

PUB SendDCC14PacketD(address14, cmd, data)          {{Command with 14 bit address + data}}
   DCCBuffer[1]:=PREAMBLE  'set preamble length
   DCCBuffer[2]:=%11_000000 | ((address14 >> 8) & %0011_1111) ' address msb 6 bits
   DCCBuffer[3]:=(address14 & $FF)                            ' address lsb 8 bits
   DCCBuffer[4]:=cmd                                        ' command
   DCCBuffer[5]:=data
   DCCBuffer[6]:=DCCBUffer[2]^DCCBuffer[3]^DCCBuffer[4]^DCCBuffer[5] ' error check
   DCCBuffer[0]:=5    'finally, set the length of the packet
   repeat until DCCBuffer[0]==0  'loop until packet is grabbed by driver

PUB SendDCC14PacketDD(address14, cmd, data1, data2) {{Command with 14 bit address + 2 data bytes}}
   DCCBuffer[1]:=PREAMBLE  'set preamble length
   DCCBuffer[2]:=%11_000000 | ((address14 >> 8) & %0011_1111) ' address msb 6 bits
   DCCBuffer[3]:=(address14 & $FF)                            ' address lsb 8 bits
   DCCBuffer[4]:=cmd                                        ' command
   DCCBuffer[5]:=data1
   DCCBuffer[6]:=data2
   DCCBuffer[7]:=DCCBUffer[2]^DCCBuffer[3]^DCCBuffer[4]^DCCBuffer[5]^DCCBuffer[6] ' error check
   DCCBuffer[0]:=6    'finally, set the length of the packet  (last 3 bytes are zero)
   repeat until DCCBuffer[0]==0  'loop until packet is grabbed by driver


PUB DccLoop | config, train, address, speed, funcs, tmp {{ COG: Sends DCC packets to all trains. Assumes 14 bit addressing, 128 speed steps }}
   repeat
      repeat train from 0 to NTRAINS-1
         tmp     := train * TRAIN_NENTRIES
         address := trains[tmp+TRAIN_ADDRESS]
         speed   := trains[tmp+TRAIN_SPEED]
         if speed < 0
            ' dcc speed is sign bit + speed. Convert 2's complement to this form
            speed := (-speed)     ' reverse direction: forward bit off
         else
            speed := $80 + speed ' forward direction: forward bit on
         funcs   := trains[tmp+TRAIN_FUNCTIONS]
         config  := trains[tmp+TRAIN_CONFIG]
         case config
            0:
               ' send speed to this train
               SendDCC7PacketD(address, %001_11111, speed)
               ' send f0-4. F1 - F4 on bits 0..3, FL on bit 4
               tmp := ((funcs & %11110) >> 1) | ((funcs & 1) << 4)
               SendDCC7Packet(address, %100_00000 + tmp)
               ' send f5-f8
               tmp := (funcs >> 5) & %1111
               SendDCC7Packet(address, %101_10000 + tmp)
               ' send f9-12
               tmp := (funcs >> 9) & %1111
               SendDCC7Packet(address, %101_00000 + tmp)
               ' send f13-f20
               tmp := (funcs >> 13) & %1111_1111
               SendDCC7PacketD(address, %110_11110, tmp)
            1:
               ' send speed to this train
               SendDCC14PacketD(address, %001_11111, speed)
               ' send f0-4
               tmp := ((funcs & %11110) >> 1) | ((funcs & 1) << 4)
               SendDCC14Packet(address, %100_00000 + tmp)
               ' send f5-f8
               tmp := (funcs >> 5) & %1111
               SendDCC14Packet(address, %101_10000 + tmp)
               ' send f9-12
               tmp := (funcs >> 9) & %1111
               SendDCC14Packet(address, %101_00000 + tmp)
               ' send f13-f20
               tmp := (funcs >> 13) & %1111_1111
               SendDCC14PacketD(address, %110_11110, tmp)

DAT
   spaces_str   BYTE "   ", 0
   am_str       BYTE "am", 0
   pm_str       BYTE "pm", 0
   welcome1_str BYTE "   WELCOME TO THE",0
   welcome2_str BYTE "     KINGSBORO",0
   welcome3_str BYTE "    SUBDIVISION", 0
   welcome4_str BYTE "Press a key to begin",0'
   any_key_str  BYTE "Press a key",0
   config_str   BYTE "Configuration Mode",0
   progrm_str   BYTE "Programming Mode",0
   variable_str BYTE "Config Var?? ",0
   prg_act_str  BYTE "Set/Query/eXit?? ",0
   value_str    BYTE "Var      Value?? ",0
   prgw_act_str BYTE "Write/Cancel/eXit?? ",0
   op1_str      BYTE "Operating",0
   op2_str      BYTE "Mode",0
   DCCBuffer    BYTE 0,0,0,0,0,0,0,0,0,0   ' 10 byte DCC packet buffer
